/*
 * Copyright 2000-2009 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.intellij.ide.actions;

import com.intellij.ide.IdeBundle;
import com.intellij.openapi.actionSystem.*;
import com.intellij.openapi.util.Key;
import com.intellij.util.ObjectUtils;
import com.intellij.util.containers.JBIterable;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;

abstract class WeighingActionGroup extends ActionGroup implements ActionWithDelegate<ActionGroup> {

  public static final Key<Double> WEIGHT_KEY = Key.create("WeighingActionGroup.WEIGHT");
  public static final Double DEFAULT_WEIGHT = Presentation.DEFAULT_WEIGHT;
  public static final Double HIGHER_WEIGHT = Presentation.HIGHER_WEIGHT;

  @Override
  public abstract @NotNull ActionGroup getDelegate();

  @Override
  public @NotNull ActionUpdateThread getActionUpdateThread() {
    return getDelegate().getActionUpdateThread();
  }

  @Override
  public void update(@NotNull AnActionEvent e) {
    getDelegate().update(e);
  }

  @Override
  public AnAction @NotNull [] getChildren(@Nullable AnActionEvent e) {
    return getDelegate().getChildren(e);
  }

  @Override
  public @NotNull List<@NotNull AnAction> postProcessVisibleChildren(@NotNull AnActionEvent e,
                                                                     @NotNull List<? extends @NotNull AnAction> visibleChildren) {
    LinkedHashSet<AnAction> heaviest = null;
    double maxWeight = DEFAULT_WEIGHT.doubleValue();
    for (AnAction action : visibleChildren) {
      Presentation presentation = e.getUpdateSession().presentation(action);
      if (!presentation.isEnabled() || !presentation.isVisible()) {
        continue;
      }
      double weight = ObjectUtils.notNull(presentation.getClientProperty(WEIGHT_KEY), DEFAULT_WEIGHT).doubleValue();
      if (weight > maxWeight) {
        maxWeight = weight;
        heaviest = new LinkedHashSet<>();
      }
      if (weight == maxWeight && heaviest != null) {
        heaviest.add(action);
      }
    }

    if (heaviest == null) {
      return new ArrayList<>(visibleChildren);
    }

    ArrayList<AnAction> chosen = new ArrayList<>();
    boolean prevSeparator = true;
    for (AnAction action : visibleChildren) {
      final boolean separator = action instanceof Separator;
      if (separator && !prevSeparator) {
        chosen.add(action);
      }
      prevSeparator = separator;

      if (shouldBeChosenAnyway(action)) {
        heaviest.add(action);
      }

      if (heaviest.contains(action)) {
        chosen.add(action);
      }
    }
    ActionGroup other = new ExcludingActionGroup(getDelegate(), heaviest);
    other.getTemplatePresentation().setText(IdeBundle.messagePointer("action.presentation.WeighingActionGroup.text"));
    other.getTemplatePresentation().setPopupGroup(true);
    return JBIterable.from(chosen).append(new Separator()).append(other).toList();
  }

  protected boolean shouldBeChosenAnyway(@NotNull AnAction action) {
    return false;
  }

}
